package de.pzzz.vertx;

import java.nio.file.Path;
import java.util.Collection;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import io.vertx.core.CompositeFuture;
import io.vertx.core.Future;
import io.vertx.core.Promise;
import io.vertx.core.Vertx;
import io.vertx.core.file.FileSystem;
import io.vertx.core.json.Json;

public class PersistentRestDataAccess<T extends SerializableWithId> extends RestDataAccess<T> {
    private static final Logger LOG = Logger.getLogger(PersistentRestDataAccess.class.getName());

    private final String baseDir;
    private final Vertx vertx;
    private boolean initialized = false;

    public PersistentRestDataAccess(final Class<T> classReference, final Vertx vertx, final String baseDir) {
        super(classReference);
        this.vertx = vertx;
        this.baseDir = baseDir;
    }

    @Override
    public void add(final T data) {
        checkInitialization();
        super.add(data);
        vertx.fileSystem().writeFile(filePath(data.getId()), Json.encodeToBuffer(data))
            .onFailure(error -> LOG.log(Level.SEVERE, "Data integrity issue! " + error.getMessage(), error));
    }

    @Override
    public void delete(final String id) {
        checkInitialization();
        super.delete(id);
        vertx.fileSystem().delete(filePath(id))
            .onFailure(error -> LOG.log(Level.SEVERE, "Data integrity issue! " + error.getMessage(), error));
    }

    @Override
    public boolean contains(String id) {
        checkInitialization();
        return super.contains(id);
    }

    @Override
    public T get(String id) {
        checkInitialization();
        return super.get(id);
    }

    @Override
    public Collection<T> list() {
        checkInitialization();
        return super.list();
    }

    @Override
    public void update(final T newData) {
        checkInitialization();
        super.update(newData);
        vertx.fileSystem().writeFile(filePath(newData.getId()), Json.encodeToBuffer(newData))
            .onFailure(error -> LOG.log(Level.SEVERE, "Data integrity issue! " + error.getMessage(), error));
    }

    public Future<PersistentRestDataAccess<T>> initialize() {
        LOG.info("Initializing " + classReference.getName() + " data access from directory " + baseDir);
        Promise<PersistentRestDataAccess<T>> promise = Promise.promise();
        FileSystem fs = vertx.fileSystem();
        fs.readDir(baseDir)
            .compose(this::readAllFiles)
            .onSuccess(ar -> {
                initialized = true;
                promise.complete(this);
            }).onFailure(promise::fail);
        return promise.future();
    }

    private void checkInitialization() {
        if (!initialized) {
            throw new IllegalStateException("PersistentRestDataAccess needs to be initialized first!");
        }
    }

    private CompositeFuture readAllFiles(final List<String> filePaths) {
        LOG.info("Found " + filePaths.size() + " files in directory " + baseDir);
        return Future.join(filePaths.stream().map(this::readFileContent).toList());
    }

    private Future<Void> readFileContent(final String filePath) {
        Promise<Void> readFuture = Promise.promise();
        vertx.fileSystem().readFile(filePath)
            .onComplete(ar -> {
                super.add(Json.decodeValue(ar.result(), classReference));
                readFuture.complete();
            })
            .onFailure(readFuture::fail);
        return readFuture.future();
    }

    private final String filePath(final String id) {
        return Path.of(baseDir, id + ".json").toString();
    }
}
