package de.pzzz.vertx.process;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

import de.pzzz.vertx.worker.QueueProcessingStatus;
import de.pzzz.vertx.PersistentRestDataAccess;
import de.pzzz.vertx.Startup;
import io.vertx.core.Future;
import io.vertx.core.Promise;
import io.vertx.core.Vertx;
import io.vertx.core.json.Json;
import io.vertx.ext.web.Router;
import io.vertx.ext.web.RoutingContext;

public abstract class ProcessController<T extends ExecutableProcess<V,U>, U extends ItemProcess<V>,V extends Serializable, W> extends PersistentRestDataAccess<T> {

    private Map<String, ProcessExecutionController<T,U,V,W>> executors = new HashMap<>();
	private final Vertx vertx;
    private final Startup startup;

    public ProcessController(final Class<T> classReference, final Vertx vertx, final Startup startup, final String baseDir) {
        super(classReference, vertx, baseDir);
        this.vertx = vertx;
        this.startup = startup;
    }

    @Override
    public void delete(final String id) {
        //TODO work with Future and improve error handling
        if (executors.containsKey(id)) {
            ProcessExecutionController<T,U,V,W> executor = executors.get(id);
            executor.stopProcessing();
            executor.close(null);
        }
        super.delete(id);
    }

    @Override
    protected void update(final String id, final RoutingContext context) {
        if (!get(id).getStatus().equals(ProcessStatus.READY)) {
            context.fail(405);
            return;
        }
        //TODO work with Future and improve error handling
        if (executors.containsKey(id)) {
            ProcessExecutionController<T,U,V,W> executor = executors.get(id);
            executor.stopProcessing();
            executor.close(null);
        }
        T newData = getDataFromRequest(context);
        newData.setId(id);
        update(newData);
    }

    @Override
    public void registerRoutes(final Router router, final String baseUrl) {
        super.registerRoutes(router, baseUrl);
        router.post(baseUrl + "/:id/start").handler(this::startProcessing);
        router.get(baseUrl + "/:id/status").handler(this::processStatus);
        router.post(baseUrl + "/:id/stop").handler(this::stopProcessing);
        router.post(baseUrl + "/:id/clear").handler(this::clearProcessingQueue);
    }

    abstract protected ProcessExecutionController<T,U,V,W> createNewProcessExecutionController(final T process, final Vertx vertx, final Startup startup, final String id);

	private void startProcessing(final RoutingContext context) {
		if (context.response().ended()) {
			return;
		}
        final String id = context.pathParam("id");
        if (!contains(id)) {
            context.fail(404);
            return;
        }
        startOrContinueProcessing(id)
            .onSuccess(status -> context.response().end(Json.encodePrettily(status)))
            .onFailure(error -> context.fail(500));
    }

    private Future<QueueProcessingStatus<U>> startOrContinueProcessing(final String id) {
        Promise<QueueProcessingStatus<U>> promise = Promise.promise();
        if (!executors.containsKey(id)) {
            T process = get(id);
            process.setStatus(ProcessStatus.RUNNING);
            ProcessExecutionController<T,U,V,W> executor = createNewProcessExecutionController(process, vertx, startup, id);
            executors.put(id, executor);
            executor.deployWorkers()
                .onSuccess(res -> promise.complete(executor.startProcessing()))
                .onFailure(error -> promise.fail(error));
        } else {
            ProcessExecutionController<T,U,V,W> executor = executors.get(id);
            promise.complete(executor.continueProcessing());
        }
        return promise.future();
    }

	private void processStatus(final RoutingContext context) {
		if (context.response().ended()) {
			return;
		}
        final String id = context.pathParam("id");
        if (!contains(id)) {
            context.fail(404);
            return;
        }
        if (!executors.containsKey(id)) {
            context.fail(400);
            return;
        }
        ProcessExecutionController<T,U,V,W> executor = executors.get(id);
		context.response().end(Json.encodePrettily(executor.processingStatus()));
    }

	private void stopProcessing(final RoutingContext context) {
		if (context.response().ended()) {
			return;
		}
        final String id = context.pathParam("id");
        if (!contains(id)) {
            context.fail(404);
            return;
        }
        if (!executors.containsKey(id)) {
            context.fail(400);
            return;
        }
        ProcessExecutionController<T,U,V,W> executor = executors.get(id);
		context.response().end(Json.encodePrettily(executor.stopProcessing()));
    }

	private void clearProcessingQueue(final RoutingContext context) {
		if (context.response().ended()) {
			return;
		}
        final String id = context.pathParam("id");
        if (!contains(id)) {
            context.fail(404);
            return;
        }
        if (!executors.containsKey(id)) {
            context.fail(400);
            return;
        }
        ProcessExecutionController<T,U,V,W> executor = executors.get(id);
		context.response().end(Json.encodePrettily(executor.clearQueue()));
    }
}
