package de.pzzz.vertx.worker;

import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.pzzz.vertx.Startup;
import io.vertx.core.DeploymentOptions;
import io.vertx.core.Future;
import io.vertx.core.Promise;
import io.vertx.core.ThreadingModel;
import io.vertx.core.Vertx;
import io.vertx.core.eventbus.DeliveryOptions;
import io.vertx.core.json.JsonObject;

public abstract class WorkerController<T, U> implements DeployableWorker {
	private static final Logger LOG = Logger.getLogger(WorkerController.class.getName());
	public static final String WORKER_BUS_ADDRESS_CONFIG = "worker.bus.address";

    private final String uuid = UUID.randomUUID().toString();
	protected final int maxWorkers;
	private String deploymentId;
	private final Vertx vertx;
	private final DeliveryOptions deliveryOptions = new DeliveryOptions();

	protected WorkerController(final int maxWorkers, final Vertx vertx, final Long timeout) {
		this.maxWorkers = maxWorkers;
		this.vertx = vertx;
		if (null != timeout) {
			deliveryOptions.setSendTimeout(timeout);
		}
	}

	public Future<Startup> deployWorkers(final Startup startup) {
		if (null != deploymentId) {
			throw new IllegalStateException("Verticle already running!");
		}
		Promise<Startup> promise = Promise.promise();
		JsonObject config = startup.getConfig();
		config.put(WORKER_BUS_ADDRESS_CONFIG, busAddress());
		LOG.info("Deploy " + maxWorkers + " workers for bus " + busAddress());
		DeploymentOptions workerOpts = new DeploymentOptions()
				.setThreadingModel(ThreadingModel.WORKER)
				.setConfig(startup.getConfig())
				.setWorkerPoolName(busAddress())
				.setInstances(maxWorkers)
				.setWorkerPoolSize(maxWorkers);
		vertx.deployVerticle(workerVerticleClass(), workerOpts)
				.onSuccess(res -> {
					deploymentId = res;
					promise.complete(startup);
				})
				.onFailure(promise::fail);
		return promise.future();
	}

	public Future<Void> undeployWorkers() {
		if (null == deploymentId) {
			throw new IllegalStateException("Verticle is not running!");
		}
		Promise<Void> promise = Promise.promise();
		vertx.undeploy(deploymentId)
			.onSuccess(promise::complete)
			.onFailure(promise::fail);
		return promise.future();
	}

	public Future<Object> doJob(final T job) {
		Promise<Object> result = Promise.promise();
		vertx.eventBus()
			.request(busAddress(), getRequest(job), deliveryOptions)
			.onSuccess(res -> result.complete(res.body()))
			.onFailure(error -> {
				LOG.log(Level.WARNING, error.getMessage(), error);
				result.fail(error);
			});
		return result.future();
	}

	protected abstract Class<? extends WorkerVerticle<U>> workerVerticleClass();

    protected String busAddress() {
        return workerVerticleClass().getName() + "." + uuid;
    }

	protected abstract U getRequest(final T job);
}
